// Sample solutions for NUMA aware loop parallelism
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <omp.h>

// Boundary value at the LHS of the bar
#define LEFT_VALUE 1.0
// Boundary value at the RHS of the bar
#define RIGHT_VALUE 10.0
// The maximum number of iterations
#define MAX_ITERATIONS 100000
// How often to report the norm
#define REPORT_NORM_PERIOD 100

int main(int argc, char * argv[]) {
	int nx, ny;	
	struct timeval start_time, end_time;

	if (argc != 5) {
		fprintf(stderr, "You must provide four command line arguments, the global size in X, the global size in Y, convergence accuracy and max number iterations\n");
		return -1;
	}
	nx=atoi(argv[1]);
	ny=atoi(argv[2]);
	double convergence_accuracy=atof(argv[3]);
	int max_its=atoi(argv[4]);

	printf("Global size in X=%d, Global size in Y=%d\n\n", nx, ny);

	int mem_size_x=nx+2;
	int mem_size_y=ny+2;

	double * u_k = malloc(sizeof(double) * mem_size_x * mem_size_y);
	double * u_kp1 = malloc(sizeof(double) * mem_size_x * mem_size_y);
	double * temp;

	double rnorm=0.0, bnorm=0.0, norm;

	int i,j,k;

#pragma omp parallel private(i,k)
	{
#pragma omp master
	{
		printf("Number of threads: %d\n", omp_get_num_threads());
	}
#pragma omp for schedule(static) private(i)
        for (i=0;i<nx+1;i++) {
                u_k[i*(ny+2)]=LEFT_VALUE;
                u_k[(ny+1)+(i*(ny+2))]=RIGHT_VALUE;
        }
#pragma omp for schedule(static) private(j,i) collapse(2)
        for (j=0;j<=nx+1;j++) {
                for (i=1;i<=ny;i++) {
                        u_k[i+(j*(ny+2))]=0.0;
                }
        }
#pragma omp for schedule(static) private(j,i) collapse(2)
        for (j=0;j<=nx+1;j++) {
                for (i=0;i<=ny+1;i++) {
                        u_kp1[i+(j*(ny+2))]=u_k[i+(j*(ny+2))];
                }
        }
#pragma omp for schedule(static) private(j,i) reduction(+:bnorm) collapse(2) 
	// Calculate the initial residual norm
	for (j=1;j<=nx;j++) {
		for (i=1;i<=ny;i++) {
			bnorm=bnorm+pow(u_k[i+(j*mem_size_y)]*4-u_k[(i-1)+(j*mem_size_y)]-
					u_k[(i+1)+(j*mem_size_y)]-u_k[i+((j-1)*mem_size_y)]-u_k[i+((j+1)*mem_size_y)], 2);
		}
	}
#pragma omp single
		{ bnorm=sqrt(bnorm); }
#pragma omp master
	{
		gettimeofday(&start_time, NULL);
	}
	for (k=0;k<MAX_ITERATIONS;k++) {
		// Calculates the current residual norm
#pragma omp for schedule(static) private(j,i) reduction(+:rnorm) collapse(2)
		for (j=1;j<=nx;j++) {
			for (i=1;i<=ny;i++) {
				rnorm=rnorm+pow(u_k[i+(j*mem_size_y)]*4-u_k[(i-1)+(j*mem_size_y)]-
						u_k[(i+1)+(j*mem_size_y)]-u_k[i+((j-1)*mem_size_y)]-u_k[i+((j+1)*mem_size_y)], 2);
			}
		}

		// Do the Jacobi iteration
#pragma omp for schedule(static) private(j,i) collapse(2)
		for (j=1;j<=nx;j++) {
			for (i=1;i<=ny;i++) {
				u_kp1[i+(j*mem_size_y)]=0.25 * (u_k[(i-1)+(j*mem_size_y)]+u_k[(i+1)+(j*mem_size_y)]+
						u_k[i+((j-1)*mem_size_y)]+u_k[i+((j+1)*mem_size_y)]);
			}
		}
#pragma omp single
		{
			// Swap data structures round for the next iteration
			temp=u_kp1;
			u_kp1=u_k;
			u_k=temp;

			norm=sqrt(rnorm)/bnorm;
			rnorm=0.0;
		}
		if (norm < convergence_accuracy) break;
		if (max_its > 0 && k >= max_its) break;
#pragma omp master
		{
			if (k % REPORT_NORM_PERIOD == 0) printf("Iteration= %d Relative Norm=%e\n", k, norm);
		}
	}
#pragma omp master
	{
		gettimeofday(&end_time, NULL);
		long int elapsedtime = (end_time.tv_sec * 1000000 + end_time.tv_usec) - (start_time.tv_sec * 1000000 + start_time.tv_usec);
		printf("\nTerminated on %d iterations, Relative Norm=%e, Total time=%ld ns\n", k, norm, elapsedtime);
	}
	}
	free(u_k);
	free(u_kp1);
	return 0;
}
